// Entwicklungsprogramm f�r den Drucksensor HP03SA von HopeRF
// lie�t Druck [hPa*10], Temp[C*10]. und errechnet die H�he [dm]
// das Programm ist lauff�hig auf dem PICkit3 extention board
// erstellt 23. August 2010

/** C O N F I G U R A T I O N   B I T S ******************************/

#pragma config FOSC = INTIO67, FCMEN = OFF, IESO = OFF                      // CONFIG1H
#pragma config PWRT = OFF, BOREN = OFF, BORV = 30                           // CONFIG2L
#pragma config WDTEN = OFF, WDTPS = 32768                                    // CONFIG2H
#pragma config MCLRE = ON, LPT1OSC = OFF, PBADEN = ON, CCP2MX = PORTC       // CONFIG3H
#pragma config STVREN = ON, LVP = OFF, XINST = OFF                          // CONFIG4L
#pragma config CP0 = OFF, CP1 = OFF, CP2 = OFF, CP3 = OFF                   // CONFIG5L
#pragma config CPB = OFF, CPD = OFF                                         // CONFIG5H
#pragma config WRT0 = OFF, WRT1 = OFF, WRT2 = OFF, WRT3 = OFF               // CONFIG6L
#pragma config WRTB = OFF, WRTC = OFF, WRTD = OFF                           // CONFIG6H
#pragma config EBTR0 = OFF, EBTR1 = OFF, EBTR2 = OFF, EBTR3 = OFF           // CONFIG7L
#pragma config EBTRB = OFF                                                  // CONFIG7H

/** I N C L U D E S **************************************************/
#include "p18f45k20.h"
#include "HP03.h"  // header file


/** V A R I A B L E S *************************************************/
#pragma udata   // declare statically allocated uinitialized variables
_HP03SA HP03SA;

// aus hp03_code.pdf
rom long Tab_BasicAltitude[80]={ // 80x 4 byte = 320

	-6983,-6201,-5413,-4620,-3820,-3015,-2203,-1385,-560, 270, //0.1m
	// 1100 1090 1080 1070 1060 1050 1040 1030 1020 1010 hpa
	1108, 1953, 2805, 3664, 4224, 5403, 6284, 7172, 8068, 8972,
	// 1000  990   980   970 960 950 940 930 920 910
	9885, 10805,11734,12671,13617,14572,15537,16510,17494,18486, 
	// 900 890 880 870 860 850 840 830 820 810
	19489,20502,21526,22560,23605,24662,25730,26809,27901,29005, 
        // 800 790 780 770 760 750 740 730 720 710
	30121,31251,32394,33551,34721,35906,37106,38322,39553,40800, 
         // 700 690 680 670 660 650 640 630 620 610
	42060,43345,44644,45961,47296,48652,50027,51423,52841,54281, 
	// 600 590 580 570 560 550 540 530 520 510 
	55744,57231,58742,60280,61844,63436,65056,66707,68390,70105, 
	// 500 490 480 470 460 450 440 430 420 410 
	71854,73639,75461,77323,79226,81172,83164,85204,87294,89438};
	// 400 390 380 370 360 350 340 330 320 310 
			 
/** P R O T O T Y P E S ******************************************************/
void InterruptServiceHigh(void);
void InterruptServiceLow(void);

/** I N T E R R U P T S ***********************************************/

//----------------------------------------------------------------------------
// High priority interrupt vector

#pragma code InterruptVectorHigh = 0x08
void InterruptVectorHigh (void)
{
  _asm
    goto InterruptServiceHigh //jump to interrupt routine
  _endasm
}

//----------------------------------------------------------------------------
// Low priority interrupt vector

#pragma code InterruptVectorLow = 0x18
void InterruptVectorLow (void)
{
  _asm
    goto InterruptServiceLow //jump to interrupt routine
  _endasm
}

/** D E C L A R A T I O N S *******************************************/
#pragma code    // declare executable instructions

// I hate delays
void delay_40_ms(void)
{
	unsigned int delay;
	for(delay =0; delay <5500; delay++);
}
//
void I2C_start(void)
{    // Wait until I2C Bus is free
 	while ((SSPCON2 & 0b00011111) | (SSPSTATbits.R_W)); // SEN, RSEN, PEN, RCEN, ACKEN;
	// dann los
 	SSPCON2bits.SEN = 1;
  	while (SSPCON2bits.SEN);	
}
//
unsigned char I2C_read(void)
{
  	SSPCON2bits.RCEN = 1;
	while(!SSPSTATbits.BF);			//while(SSPCON2bits.RCEN);
  	return SSPBUF;
} 
//
unsigned char I2C_write (unsigned char value)
{
	PIR1bits.SSPIF = 0;
  	SSPBUF = value;					
  	while(!PIR1bits.SSPIF);			
	PIR1bits.SSPIF = 0;			
	return SSPCON2bits.ACKSTAT;
}
void I2C_master_ACK(unsigned char yes_no)
{
	 // Master acknowlage
	SSPCON2bits.ACKDT = yes_no;
	SSPCON2bits.ACKEN = 1;
  	while(SSPCON2bits.ACKEN);	
}
//
void I2C_stop(void)
{
	SSPCON2bits.PEN = 1;   		// stop
	while(SSPCON2bits.PEN);
}
// "Roh_Druck und Temp" von HP03SA lesen
// setzt eine Initialisierung der Messung voraus
// das pasiert im ersten Schritt dann 40 ms warten dann lesen
WORD HP03SA_read_value(unsigned char code)
{ WORD help;
	
	HP03SA_XCLR = hight;
		
	I2C_start();
	I2C_write(HP03_write_addr);	// 0xEE	
	I2C_write(0xFF);		// 0xFF
	I2C_write(code);		// 0xE8 = Temperatur
					// 0xF0 = Druck	
	I2C_stop();
	delay_40_ms();                  // conversation
	 // conversion done read the values
	I2C_start();
	I2C_write(HP03_write_addr);	// 0xEE
	I2C_write(0xFD);		// 0xFD
		
	I2C_start();				
	I2C_write(HP03_read_addr);	// 0xEF
						
	help.HighB = I2C_read();
	I2C_master_ACK(yes);		// weiteren Wert lesen
		
	help.LowB = I2C_read();		
	I2C_master_ACK(no);		// letzter Wert
		
	I2C_stop();

	HP03SA_XCLR = low;
	return help;
}

// um den wirklichen Druckwert zu ermitteln
// brauchen wir die Parameter aus dem Sensor
// einmalig bei jedem boot; sind im EEPROM des
// Sensors abgelegt hole die Daten nach HP03SA.xxx
void HP03SA_Koeffizienten_lesen(void)
{ unsigned char dummy;
	
	//Make sure to pull low XCLR before start to Read these coefficients or the data is probably incorrect
	HP03SA_XCLR = low;
		
	I2C_start();
	I2C_write(HP03_EEPROM_addr_write); 	// 0xA0
	I2C_write(16);
	I2C_start();	
	I2C_write(HP03_EEPROM_addr_read); 	// 0xA1
// C1
 	HP03SA.wC1.HighB = I2C_read();
	I2C_master_ACK(yes); 	
 	HP03SA.wC1.LowB = I2C_read();
	I2C_master_ACK(yes); 	
// C2	
 	HP03SA.wC2.HighB = I2C_read();
	I2C_master_ACK(yes);
 	HP03SA.wC2.LowB = I2C_read();
	I2C_master_ACK(yes); 	
// C3 	
 	HP03SA.wC3.HighB = I2C_read();
	I2C_master_ACK(yes); 	
 	HP03SA.wC3.LowB = I2C_read();
	I2C_master_ACK(yes); 	
// C4 	
 	HP03SA.wC4.HighB = I2C_read();
	I2C_master_ACK(yes); 	
 	HP03SA.wC4.LowB = I2C_read();
	I2C_master_ACK(yes); 	
// C5 	
 	HP03SA.wC5.HighB = I2C_read();
	I2C_master_ACK(yes); 	
 	HP03SA.wC5.LowB = I2C_read();
	I2C_master_ACK(yes); 	
// C6 	
 	HP03SA.wC6.HighB = I2C_read();
	I2C_master_ACK(yes); 	
 	HP03SA.wC6.LowB = I2C_read();
	I2C_master_ACK(yes); 	
// C7 	
 	HP03SA.wC7.HighB = I2C_read();
	I2C_master_ACK(yes); 	
 	HP03SA.wC7.LowB = I2C_read();
	I2C_master_ACK(yes); 	
// AA
        HP03SA.AA = I2C_read();
	I2C_master_ACK(yes);
// BB	
        HP03SA.BB = I2C_read();
	I2C_master_ACK(yes);
// CC	
        HP03SA.CC = I2C_read();
	I2C_master_ACK(yes);
// DD	
        HP03SA.DD = I2C_read();
	I2C_master_ACK(no);
 	
	I2C_stop();
}
// 
void HP03SA_Druck_Temp_Hoehe_errechnen(void)
{
	signed long dUT, SENS, X, help;
	float OFF, frac;
	unsigned int faktor, basicpressure, pressure, altitude;
	unsigned char  i;
	//long druck_klein, druck_aktuell, differenz;

	help = (HP03SA.D2 -HP03SA.C5);

	if (help < 0)	faktor = HP03SA.BB;		// D2 < C5
	else 			faktor = HP03SA.AA;	// D2 >= C5

	dUT = help -((faktor *help *help / 16384) >> HP03SA.CC); // -5478 tested ok
	
	OFF = (HP03SA.C4 -1024) *dUT;
	OFF /= 16384;
	OFF += HP03SA.C2;
	OFF *= 4;												// 15675 tested ok
	
	SENS = HP03SA.C1 +((HP03SA.C3* dUT)/1024);	// 28238 tested ok
	X = (HP03SA.D1 -7168);
	X *= SENS;
	X /= 16384;
	X -= OFF;												// 23737 tested ok (rundungsfehler)
// Druck
	HP03SA.Druck_hPa = (X*100/32) + (HP03SA.C7 *10);    // 99100 - 991,00 hPa tested ok
// Temp
	HP03SA.Temp_C	= 250 + ((dUT * HP03SA.C6)/65536) -(dUT >> HP03SA.DD);

// Höhe neu
	for (i = 0; ; i++) 
	{
		basicpressure = (1100 -i*10);
		if (basicpressure < HP03SA.Druck_hPa/100) break;
	};
   	pressure = HP03SA.Druck_hPa - basicpressure*100;
   	altitude = (Tab_BasicAltitude[i] - Tab_BasicAltitude[i-1]) *pressure/1000;
   	HP03SA.Hoehe_dm = Tab_BasicAltitude[i] -altitude;
   	
   	i = (HP03SA.Hoehe_dm %10);	// four lose and five up
	if (HP03SA.Hoehe_dm < 0)   	
	{
		if (i > 4)
			HP03SA.Hoehe_dm -= 10 -i;
		else
			HP03SA.Hoehe_dm += i;	
	}
	else 
	{
		if (i > 4)
			HP03SA.Hoehe_dm += 10 -i;
		else
			HP03SA.Hoehe_dm -= i;			
	}		
 }


void main (void)
{
// OSCI
	OSCCON = 0b01100010;		// interer Oscillator 8 MHz
// analog
	ANSELH = 0;
// Init I/O
	PORTC = 0;
	TRISCbits.TRISC6 = 0;		// XCLR = RC6 - output
	HP03SA_XCLR = low;		// PORTCbits.RC6
// I2C - max. 500 kHz f�r Sensor
	TRISCbits.TRISC4 = 1;
	TRISCbits.TRISC3 = 1;
	SSPCON2 = 0x00;
	SSPADD = 19;			// 19 = 100 kHz Baud Rate in master mode
	SSPCON1 = 0b00111000;		// SSPEN, CKP, SSPM3
	#ifdef __debug
	// um im debug mode SCL und SDA releasen
	SSPCON1bits.SSPEN = 0;	
	SSPCON1bits.SSPEN = 1;	
	#endif	
// PWM clock output for the pressure sensor HP03
	TRISCbits.TRISC1 = 0; 		// MCLK = RC1 - output
	CCP2CON = 0b00001100;		// PWM mode
	CCPR2L = 30;			// 50% duty cycle
	PR2 = 60;			// 32,89 kHz
	T2CONbits.TMR2ON = 1;
//INT0 switch
	TRISBbits.TRISB0 = 1;
        INTCON2bits.INTEDG0 = 0;
        INTCONbits.INT0IF = 0;      // ensure flag is cleared
        INTCONbits.INT0IE = 1;      // enable INT0 interrupt
// Set up global interrupts
         RCONbits.IPEN = 0;         // Enable priority levels on interrupts
         INTCONbits.GIEL = 1;       // Low priority interrupts allowed
        INTCONbits.GIEH = 1;        // Interrupting enabled.
//
	HP03SA_Koeffizienten_lesen();
	
    while (1);			// nothing just wait interrupts (I like it)
}

// -------------------- Iterrupt Service Routines --------------------------
//#pragma interrupt InterruptServiceHigh nosave= TBLPTR, TBLPTRU, TABLAT, PCLATH, PCLATU, PROD, FSR0
#pragma interrupt InterruptServiceHigh 
void InterruptServiceHigh(void)
{	
	// Taster gedrückt
	if (INTCONbits.INT0IF)
	{
    	HP03SA.wD1 = HP03SA_read_value(Roh_Druck);
    	HP03SA.wD2 = HP03SA_read_value(Roh_Temperatur);	       
	HP03SA_Druck_Temp_Hoehe_errechnen();	   	    
		
	INTCONbits.INT0IF = 0;		
	}

}  


#pragma interruptlow InterruptServiceLow// "interruptlow" pragma for low priority
void InterruptServiceLow(void)
{

}